/*
 * Process Variable Library with JVisu Socket Server
 *
 * Copyright (C) 2003 Thomas Gleixner <tglx@linutronix.de>
 *           (C) 2005 Robert Schwebel <r.schwebel@pengutronix.de>
 *           (C) 2005 Sascha Hauer <s.hauer@pengutronix.de>
 *
 * file: pv_bs_8bit.c
 *
 * Please send bug reports to bugs@pengutronix.de
 *
 * See COPYING for the status of this software.
 *
 * PURPOSE
 * This is a backing store for using 8Bit (FRAM, SRAM) devices on a
 * Processor which can't do 8Bit accesses.
 *
 * We use /dev/mem for mapping the device into memory, so we need root
 * privileges
 * 
 */

#include <libpv_config.h>

#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <pv_bs_8bit.h>


struct pv_bs_8bit_priv {
	int fd;
	void *mem;
};


static int pv_bs_8bit_exit(pv_backing_store_t *store) {
	struct pv_bs_8bit_priv *priv = store->priv;

	munmap(priv->mem, store->size);
	close(priv->fd);

	return 0;
}


static int pv_bs_8bit_read(pv_backing_store_t *store, off_t offset, size_t len) {
	struct pv_bs_8bit_priv	*priv = store->priv;
	unsigned long		*src = priv->mem + (offset << 2);
	unsigned char		*dst = store->map + offset;
	int			i;

	dprintf(LEVEL_ALL, "(%s) off %d, len %zd\n",
		__FUNCTION__, (int)offset, len);

	for (i=0; i<len; i++)
		*dst++ = *src++;

	return 0;
}


static int pv_bs_8bit_write(pv_backing_store_t *store, off_t offset, size_t len) {
	struct pv_bs_8bit_priv	*priv = store->priv;
	unsigned char		*src = store->map + offset;
	unsigned long		*dst = priv->mem + (offset << 2);	
	int i;

	dprintf(LEVEL_ALL, "%s: off: %d len: %zd\n",
		__FUNCTION__, (int)offset, len);

	for (i=0; i<len; i++)
		*dst++ = *src++;

	return 0;
}


int pv_bs_8bit_init(pv_backing_store_t	*store,
		    const char		*filename,
		    size_t		size) {
	struct pv_bs_8bit_priv	*priv;
	int			err;

	dprintf(LEVEL_INFO, "(%s), initialising 8Bit backing store\n",
		__FUNCTION__);

	priv = calloc(1, sizeof(struct pv_bs_8bit_priv));
	if(!priv) {
		err = -ENOMEM;
		dprintf(LEVEL_CRIT, "(%s) failed to alloc private mem, err %d\n",
			__FUNCTION__, err);
		goto exit;
	}

	store->priv  = priv;
	store->exit  = pv_bs_8bit_exit;
	store->read  = pv_bs_8bit_read;
	store->write = pv_bs_8bit_write;
	store->size  = size;

	err = open(filename, O_RDWR | O_SYNC);
	if(err < 0) {
		err = -errno;
		dprintf(LEVEL_CRIT, "(%s) failed to open backing store \"%s\", err %d\n",
			__FUNCTION__, filename, err);
		goto exit_free;
	}
	priv->fd = err;

	priv->mem = mmap(NULL, store->size, PROT_READ | PROT_WRITE, 
		         MAP_SHARED, priv->fd, 0);
	if(priv->mem == MAP_FAILED) {
		err = -errno;
		dprintf(LEVEL_CRIT, "(%s) failed to mmap backing store \"%s\", err %d\n",
			__FUNCTION__, filename, err);
		goto exit_close;
	}

	store->map = NULL;	/* no direct access to mem allowed */

	return 0;

 exit_close:
	close(priv->fd);
 exit_free:
	free(priv);
 exit:
	return err;
}

/* end of file pv_bs_8bit.c */
